#include "CoCoA/BuildInfo.H"
#include "CoCoA/DenseUPolyRing.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/GlobalManager.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingDenseUPolyClean.H"
#include "CoCoA/RingQ.H"
#include "CoCoA/RingZ.H"
#include "CoCoA/io.H"
#include "CoCoA/symbol.H"

using namespace CoCoA;

#include <iostream>
using std::endl;
#include <vector>
using std::vector;

//----------------------------------------------------------------------
// Test for DenseUPolyRing functions on RingElem
// functions: +, -f, f-g, *, StdDeg
// environments: DenseUPolyClean (Fp, Z)
//----------------------------------------------------------------------

#define TEST_ASSERT(cond)   do { if (!(cond)) CoCoA::AssertionFailed(#cond, __FILE__, __LINE__); } while(0)


void TestDenseUPolyRing(DenseUPolyRing P)
{
  GlobalOutput() << "TESTING: " << P << endl << endl;

  
  const RingElem& x = indet(P, 0);
  RingElem f1 = IndetPower(P, 0, 2) + 3*x +2,
    f2 = IndetPower(P, 0, 4) + x,
    f3 = x + 1;

  GlobalOutput() << "Given f1 = " << f1 << endl;
  GlobalOutput() << "  and f2 = " << f2 << endl << endl;
 
   GlobalOutput() << "f1+f2   gives  " << f1+f2 << endl;
   GlobalOutput() << "f2+f1   gives  " << f2+f1 << endl;
   GlobalOutput() << "f2-f1   gives  " << f2-f1 << endl;
   GlobalOutput() << "-f1   gives  "   << -f1 << endl;
   GlobalOutput() << "deg((2*x+1)*(3*x+1)) gives  "   << deg((2*x+1)*(3*x+1)) << endl;
//   //  GlobalOutput() << "gcd(f1,f2)   gives  " << gcd(f1,f2) << endl;
   GlobalOutput() << "StdDeg(f1)   gives  " << StdDeg(f1) << endl;
   GlobalOutput() << "StdDeg(f2)   gives  " << StdDeg(f2) << endl;

//   if ( IsField(CoeffRing(P)) )
//   {
//     TEST_ASSERT( IsInvertible(gcd(f1*f2, f1*f3)/f1) );
//     TEST_ASSERT( IsInvertible(gcd(f1*f2, f2*f3)/f2) );
//     TEST_ASSERT( IsInvertible(gcd(f1*f3, f2*f3)/f3) );
//   }
  
// Check some computations (without printing any results).
   GlobalOutput() << "one(P)  " << one(P) << endl;
   TEST_ASSERT(IsOne(one(P)));
   TEST_ASSERT(!IsOne(f1));
   TEST_ASSERT(!IsOne(f2));

   RingElem f1f2 = f1*f2;
   RingElem f2f1 = f2*f1;
   if (IsCommutative(P))  TEST_ASSERT(f1f2 == f2f1);
   TEST_ASSERT(f1 != f2);
   TEST_ASSERT(f1+f2 == f2+f1);
   TEST_ASSERT(f1+(-f1) == 0);
   TEST_ASSERT(f1-f1 == 0);
   TEST_ASSERT(f1f2 != f1);
   TEST_ASSERT(f1f2 != f2);
//   TEST_ASSERT(IsDivisible(f1f2, f1));
//   TEST_ASSERT(IsDivisible(f1f2, f2));
//   TEST_ASSERT(!IsDivisible(f1, f2));
//   TEST_ASSERT(!IsDivisible(f2, f1));
//   TEST_ASSERT(f1f2/f1 == f2);
//   TEST_ASSERT(f1f2/f2 == f1);
//   //  TEST_ASSERT(gcd(f1*f1f2, f2*f1f2) == f1f2);

//   f2f1 = 1;
//   TEST_ASSERT(IsOne(f2f1));

  GlobalOutput() << "------------------------------------------------" << endl << endl;
}


void program()
{
  GlobalManager CoCoAFoundations;

  const QuotientRing Fp = NewZmod(101);
  const QuotientRing F6 = NewZmod(6);
  const DenseUPolyRing Zx = NewPolyRing_DUP(RingZ());
  const DenseUPolyRing Fpx = NewPolyRing_DUP(Fp);
  const DenseUPolyRing F6y = NewPolyRing_DUP(F6, symbol("y"));
  const DenseUPolyRing Qx = NewPolyRing_DUP(RingQ());

  TestDenseUPolyRing(Zx);
  TestDenseUPolyRing(Fpx);
  TestDenseUPolyRing(F6y);
  TestDenseUPolyRing(Qx);
}


// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA Error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }

  BuildInfo::PrintAll(GlobalErrput());
  return 1;
}
